#include <stdio.h>
#include <stdlib.h>
#include "clut.h"

#define IDX(i,j,n) ((i)*(n)+(j))
#define DEBUG 0

#define N                   2000
#define LOCAL_SIZE          22               // work items per work group

void expimg(int* I, int* O, int n) {
    int i, j;
    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            int v = I[IDX(i,j,n)];
            O[IDX(2*i,   2*j,   2*n)] = v;
            O[IDX(2*i+1, 2*j,   2*n)] = v;
            O[IDX(2*i,   2*j+1, 2*n)] = v;
            O[IDX(2*i+1, 2*j+1, 2*n)] = v;
        }
    }
}

void printimg(int* I, int n){
    int i, j;

    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            printf("%-4d ", I[i*n+j]);
        }
        printf("\n");
    }
}

int compmat(int* A, int* B, int n){
    int i, j;
    
    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            if (A[IDX(i,j,n)] != B[IDX(i,j,n)]) return 0;
        }
    }
    return 1;
}

int main(void) {
    
    clut_device dev;	// context, device queue & program
    int       err;      // error code
    cl_kernel kernel;   // execution kernel

    // create the two input vectors + output vector
    int i, j, n = N;
    int *I = (int*)malloc(sizeof(int)*n*n);
    int *O = (int*)malloc(sizeof(int)*2*n*2*n);
    int *OCPU = (int*)malloc(sizeof(int)*2*n*2*n);

    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            I[i*n+j] = i+j;
        }
    }
    
	clut_open_device(&dev, "expimg.cl");
    
    clut_print_device_info(&dev);

    // create memory buffers on the device for each vector 
    cl_mem i_mem = clCreateBuffer(dev.context, 
    						CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 
    						n*n * sizeof(int), I, NULL);
    if (!i_mem) clut_panic("failed to allocate input img on device memory");

    cl_mem o_mem = clCreateBuffer(dev.context, CL_MEM_WRITE_ONLY, 
    						2*n*2*n * sizeof(int), NULL, NULL);
    if (!o_mem) clut_panic("failed to allocate output img on device memory");

    // create an OpenCL kernel
    kernel = clCreateKernel(dev.program, "matrix_add", &err);
    clut_check_err(err, "clCreateKernel failed");

    // set the arguments of the kernel
    err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&i_mem);
    err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&o_mem);
    err |= clSetKernelArg(kernel, 2, sizeof(int), (void *)&n);
    clut_check_err(err, "clSetKernelArg failed");

    // execute the OpenCL kernel on the list
    size_t local_item_size[2]  = { LOCAL_SIZE, LOCAL_SIZE };
    size_t global_item_size[2] = 
        { ((n+LOCAL_SIZE-1)/LOCAL_SIZE)*LOCAL_SIZE,
          ((n+LOCAL_SIZE-1)/LOCAL_SIZE)*LOCAL_SIZE } ;

    printf("data size: %d\n", n);
    printf("global size: %lu\n", global_item_size[0]);
    
    err = clEnqueueNDRangeKernel(dev.queue, kernel, 2, NULL, 
                                 global_item_size, 
                                 local_item_size, 0, NULL, NULL);
    clut_check_err(err, "clEnqueueNDRangeKernel failed");

    // read the memory buffer C on the device to the local variable C
    err = clEnqueueReadBuffer(dev.queue, o_mem, CL_TRUE, 0, 
                              2*n*2*n * sizeof(int), O, 0, NULL, NULL);
    clut_check_err(err, "clEnqueueReadBuffer failed");


    expimg(I, OCPU, n);
    
    // stampa risultato
#if DEBUG
    printf("\nMatrice I\n");
    printimg(I, n);
    
    printf("\nMatrice O CPU\n");
    printimg(OCPU, 2*n);

    printf("\nMatrice O GPU\n");
    printimg(O, 2*n);
#endif

	if (!compmat(OCPU,O,2*n)) printf("correctness test: FAILED\n");
	else                      printf("correctness test: PASSED\n");

    clut_close_device(&dev);

    free(OCPU);
    free(O);
    free(I);
    return 0;
}
