// =====================================================================
//  main.c
// =====================================================================

//  Author:         (c) 2013 Camil Demetrescu
//  License:        See the end of this file for license information
//  Created:        June 16, 2013
 
//  Last changed:   $Date: 2013/06/16 15:00:07 $
//  Changed by:     $Author: demetres $
//  Revision:       $Revision: 1.00 $


#include <stdio.h>
#include <stdlib.h>

#include "clut.h"
#include "pgm.h"
#include "mandelbrot.h"

#define HOST_DIR        "img_host"
#define DEVICE_DIR      "img_device"
#define PROG_NAME       "mandelbrot.cl"
#define IMAGE_EXT       ".pgm"
#define MAX_LEN         1024
#define DEFAULT_ITER    300
#define RMIN			-2.5
#define RMAX			1.0
#define IMIN			-1.0
#define IMAX			1.0
#define DIFF_THRESHOLD  (0.01)


// ---------------------------------------------------------------------
// test_correctness
// ---------------------------------------------------------------------
// compare the two images

void test_correctness(unsigned char* img_host, unsigned char* img_device, int w, int h) {
    int i, j;
    double sum_host = 0, sum_device = 0, avg_err;

    printf("    Correctness test:\n"); 

    for (i=0; i<h; i++)
        for (j=0; j<w; j++) {
        	sum_host   += img_host[i*w+j];
        	sum_device += img_device[i*w+j];
    	}

    avg_err = abs(sum_host-sum_device)/sum_host;

    printf("        Host image checksum = %f\n", sum_host);
    printf("        Device image checksum = %f\n", sum_device);
    printf("        Average error = %g%%\n", avg_err*100);
    printf("        >> Test %s\n", avg_err < DIFF_THRESHOLD ? "PASSED" : "FAILED ###");
}


// ---------------------------------------------------------------------
// save_image
// ---------------------------------------------------------------------
// save image to disk

void save_image(unsigned char* img, float rc, float ic, float mag, 
               int w, int h, int max_iter, char* res_dir) {
    char img_path[MAX_LEN]; // pathname of output image file

	// build pathname of output image file
	sprintf(img_path, "%s/mb_%f_%f_%f_%ux%u_%d%s", 
			res_dir, rc, ic, mag, w, h, max_iter, IMAGE_EXT);

	// save output image
	printf("Saving image %s\n", img_path);
	int err = pgm_save(img, h, w, img_path);
	if (err) clut_panic("failed to save output image file");
}


// ---------------------------------------------------------------------
// run_test
// ---------------------------------------------------------------------
// build image representing mandelbrot set
void run_test(float rc, float ic, float scale, float mag,
		      unsigned w, unsigned h, unsigned max_iter,
			  unsigned char* img, clut_device* dev) {

    double t, td;
    unsigned char *img_host, *img_device;

    // allocate correct output image to be computed by host
	img_host = malloc(w*h*sizeof(unsigned char));
	if (img_host == NULL) 
		clut_panic("failed to allocate host-computed output image on host");

    // allocate correct output image to be computed by device
	img_device = malloc(w*h*sizeof(unsigned char));
	if (img_device == NULL) 
		clut_panic("failed to allocate device-computed output image on host");

    // compute mandelbrot set on host
    mandelbrot_host(rc, ic, scale, w, h, max_iter, img_host, &t);

    // compute mandelbrot set on device
    mandelbrot_device(rc, ic, scale, w, h, max_iter, img_device, dev, &td);

    // validate results
    printf("Computing Mandelbrot set\n");
    printf("    CPU took %.3f msec\n", t*1E03);
    printf("    Device took %.3f msec\n", td*1E03);
    
    // perform correctness test
    test_correctness(img_host, img_device, w, h);

	// saving images to disk
	save_image(img_host, rc, ic, mag, w, h, max_iter, HOST_DIR);
	save_image(img_device, rc, ic, mag, w, h, max_iter, DEVICE_DIR);

    // cleanup
	free(img_host);
	free(img_device);
}


// ---------------------------------------------------------------------
// main
// ---------------------------------------------------------------------
int main(int argc, char** argv) {

    clut_device     dev;                    // device structure
    unsigned char*  img;                    // output image on host
    float           rc, ic, mag;            // mandelbrot domain
    unsigned        w, h;				    // image width and height in pixels
    unsigned        iter;				    // number of iterations

    // check number of input parameters
    if (argc < 6) {
		fprintf(stderr, "usage: ./%s rc ic mag width height [iter]\n", argv[0]);
		exit(1);
    }
    
    // get input parameters
    rc    = atof(argv[1]);
    ic    = atof(argv[2]);
    mag   = atof(argv[3]);
    w     = atoi(argv[4]);
	h     = atoi(argv[5]);
    if (argc > 6) iter = atoi(argv[6]);
	else          iter = DEFAULT_ITER;

	// trim input parameters to fit Mandelbrot region
	if (rc < RMIN) rc = RMIN;
	if (rc > RMAX) rc = RMAX;
	if (ic < IMIN) ic = IMIN;
	if (ic > IMAX) ic = IMAX;

	// compute pixel size for Mandelbrot set at magnification factor 1x
	float pixel_h = (IMAX-IMIN)/(h*mag);
	float pixel_w = (RMAX-RMIN)/(w*mag);
	float pixel_size = pixel_h < pixel_w ? pixel_h : pixel_w;

	printf("Mandelbrot set parameters:\n");
	printf("    center=(%f,%f), magnification=%fx\n", rc, ic, mag);
	printf("    width=%u, height=%u, max iterations=%u\n", w, h, iter);
	printf("    pixel_size=%g\n", pixel_size);

    // open device
    clut_open_device(&dev, PROG_NAME);

    // print device info
    clut_print_device_info(&dev);

	// run test
	run_test(rc, ic, pixel_size, mag, w, h, iter, img, &dev);
    
    // cleanup
    clut_close_device(&dev);
    
    printf("\n");
    
    return 0;
}


// Copyright (C) 2013 Camil Demetrescu
  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
  
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
  
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA
